-------------------------------------------------------------------------------
-- Title      : controlUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Control unit of a sequential FIR filter.
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2018-04-11  1.0      jnbazin Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity controlUnit is

  port (
    I_clock               : in  std_logic;  -- global clock
    I_reset               : in  std_logic;  -- asynchronous global reset
    I_inputSampleValid    : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_processingDone_1    : in std_logic;
    I_processingDone_2b   : in std_logic;
    I_processingDone_2a   : in std_logic;
    I_processingDone_3    : in std_logic;
    O_loadShift_1         : out std_logic;  -- filtered sample
    O_loadShift_2b        : out std_logic;  -- filtered sample
    O_loadShift_2a        : out std_logic;  -- filtered sample
    O_loadShift_3         : out std_logic;  -- filtered sample
    O_initAddress         : out std_logic;  -- Control signal to initialize register read address
    O_incrAddress         : out std_logic;  -- Control signal to increment register read address
    O_initSum             : out std_logic;  -- Control signal to initialize the MAC register
    O_loadSum             : out std_logic;  -- Control signal to load the MAC register;
    O_loadY               : out std_logic;  -- Control signal to load Y register
    O_FilteredSampleValid : out std_logic  -- Data valid signal for filtered sample
    );

end entity controlUnit;
architecture archi_operativeUnit of controlUnit is


  type T_state is (WAIT_SAMPLE, STORE, PROCESSING_LOOP_FIR_1, PROCESSING_LOOP_IIR_b, Init_address, PROCESSING_LOOP_IIR_a, PROCESSING_LOOP_FIR_2, OUTPUT, WAIT_END_SAMPLE);  -- state list
  signal SR_presentState : T_state;
  signal SR_futurState   : T_state;

begin

  process (I_reset, I_clock) is -- si horloge alors synchrone séquentiel
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_presentState <= WAIT_SAMPLE;
    elsif rising_edge(I_clock) then     -- rising clock edge
      SR_presentState <= SR_futurState;
    end if;
  end process;

  process (SR_presentState,I_processingDone_1, I_processingDone_2a, I_processingDone_3, I_processingDone_2b,I_inputSampleValid) is --logique combinatoire
  begin
    case SR_presentState is

      when WAIT_SAMPLE =>
        if I_inputSampleValid = '1' then
            SR_futurState <= STORE;
        else
            SR_futurState <= WAIT_SAMPLE;
        end if;
            
      when STORE =>
        SR_futurState <= PROCESSING_LOOP_FIR_1;
        
      when PROCESSING_LOOP_FIR_1 =>
        if I_processingDone_1 = '1' then
            SR_futurState <= PROCESSING_LOOP_IIR_b;
        else
            SR_futurState <= PROCESSING_LOOP_FIR_1;
        end if;
            
      when PROCESSING_LOOP_IIR_b =>
        if I_processingDone_2b = '1' then
                  SR_futurState <= Init_address;
              else
                  SR_futurState <= PROCESSING_LOOP_IIR_b;
              end if;
        
      when Init_address =>
        SR_futurState <= PROCESSING_LOOP_IIR_a;
        
      when PROCESSING_LOOP_IIR_a =>
                if I_processingDone_2a = '1' then
                          SR_futurState <= PROCESSING_LOOP_FIR_2;
                      else
                          SR_futurState <= PROCESSING_LOOP_IIR_a;
                      end if;
                      
      when PROCESSING_LOOP_FIR_2 =>
         if I_processingDone_3 = '1' then
            SR_futurState <= Output;
          else
            SR_futurState <= PROCESSING_LOOP_FIR_2;
          end if;
          
      when Output =>
        SR_futurState <= WAIT_END_SAMPLE;
        
      when WAIT_END_SAMPLE=> 
        if I_inputSampleValid = '0' then
                  SR_futurState <= WAIT_SAMPLE;
              else
                  SR_futurState <= WAIT_END_SAMPLE;
              end if;

      when others => null;
    end case;
  end process;

  O_loadShift_1         <= '1' when SR_presentState = PROCESSING_LOOP_FIR_1 else '0';
  O_loadShift_2b        <= '1' when I_processingDone_1 = '1' else '0';
  O_loadShift_2a        <= '1' when I_processingDone_2b='1' else '0';
  O_loadShift_3         <= '1' when I_processingDone_2a='1' else '0';
  O_initAddress         <= '1' when SR_presentState = STORE or SR_presentState = Init_address else '0';
  O_incrAddress         <= '1' when SR_presentState =  PROCESSING_LOOP_FIR_1 or SR_presentState =  PROCESSING_LOOP_IIR_b or SR_presentState =  PROCESSING_LOOP_IIR_a or SR_presentState =  PROCESSING_LOOP_FIR_2 else '0';
  O_initSum             <= '1' when SR_presentState = STORE else '0';
  O_loadSum             <= '1' when SR_presentState =  PROCESSING_LOOP_FIR_1 or SR_presentState =  PROCESSING_LOOP_IIR_b or SR_presentState =  PROCESSING_LOOP_IIR_a or SR_presentState =  PROCESSING_LOOP_FIR_2 else '0';
  O_loadY               <= '1' when SR_presentState = OUTPUT else '0';
  O_FilteredSampleValid <= '1' when SR_presentState = WAIT_END_SAMPLE else '0';





end architecture archi_operativeUnit;
